@@ -2,7 +2,7 @@
 
 -- 
 -- Copyright (c) 2005-2008 Don Stewart - http://www.cse.unsw.edu.au/~dons
--- Copyright (c) 2008, 2019-2024 Galen Huntington
+-- Copyright (c) 2008, 2019-2025 Galen Huntington
 -- 
 -- This program is free software; you can redistribute it and/or
 -- modify it under the terms of the GNU General Public License as
@@ -33,12 +33,10 @@ module Core (
         seekStart,
         blacklist,
         showHist, hideHist,
-        writeSt, readSt,
         jumpToMatch, jumpToMatchFile,
         toggleFocus, jumpToNextDir, jumpToPrevDir,
         loadConfig,
         discardErrors,
-        FileListSource,
         toggleExit,
     ) where
 
@@ -67,13 +65,10 @@ import System.IO                (hPutStrLn, hGetLine, stderr, hFlush)
 import System.Process           (runInteractiveProcess, waitForProcess)
 import System.Clock             (TimeSpec(..), diffTimeSpec)
 import System.Random            (randomIO)
-import System.FilePath          ((</>), takeDirectory)
-import Data.List                (isInfixOf, tails)
+import System.FilePath          ((</>))
 
 import System.Posix.Process     (exitImmediately)
 
-type FileListSource = Either SerialT [ByteString]
-
 
 mp3Tool :: String
 mp3Tool =
@@ -85,24 +80,15 @@ mp3Tool =
 
 ------------------------------------------------------------------------
 
-start :: Bool -> FileListSource -> IO ()
-start playNow ms = handle @SomeException (shutdown . Just . show) do
+start :: Bool -> Tree -> IO ()
+start playNow (Tree ds fs) = handle @SomeException (shutdown . Just . show) do
 
     t0 <- forkIO mpgLoop    -- start this off early, to give mpg123 time to settle
 
     c <- UI.start -- initialise curses
 
-    (ds,fs,i,m)   -- construct the state
-        <- case ms of
-           Right roots -> do Tree a b <- buildTree roots
-                             pure (a,b,0,Normal)
-
-           Left st     -> pure (ser_darr st
-                               ,ser_farr st
-                               ,ser_indx st
-                               ,ser_mode st)
-
     now <- getMonoTime
+    mode <- readState
 
     -- fork some threads
     t1 <- forkIO $ mpgInput readf
@@ -117,9 +103,9 @@ start playNow ms = handle @SomeException (shutdown . Just . show) do
         { music        = fs
         , folders      = ds
         , size         = 1 + (snd . bounds $ fs)
-        , cursor       = i
-        , current      = i
-        , mode         = m
+        , cursor       = 0
+        , current      = 0
+        , mode         = mode
         , uptime       = showTimeDiff now now
         , boottime     = now
         , config       = c
@@ -127,7 +113,8 @@ start playNow ms = handle @SomeException (shutdown . Just . show) do
 
     loadConfig
 
-    when (0 <= (snd . bounds $ fs)) play -- start the first song
+    when (0 <= (snd . bounds $ fs)) do
+        if mode == Random then modifySTM jumpToRandom else playCur
     when (not playNow) pause
 
     run         -- won't restart if this fails!
@@ -143,6 +130,10 @@ runForever fn = catch (forever fn) handler
             unless (exitTime e) $
                 (warnA . show) e >> runForever fn        -- reopen the catch
 
+-- | Standard combinator.
+whenJust :: Monad m => Maybe a -> (a -> m ()) -> m ()
+whenJust mb f = maybe (pure ()) f mb
+
 -- | Generic handler
 -- I don't know why these are ignored, but preserving old logic.
 -- For profiling, make sure to return True for anything:
@@ -170,32 +161,36 @@ mpgLoop = runForever do
       Just mppath -> do
 
         -- if we're never able to start mpg123, do something sensible
-        --   TODO no need for this Maybe unpacking, just catch and rerun loop
         mv <- catch (pure <$> runInteractiveProcess mppath ["-R", "-"] Nothing Nothing)
                     (\ (e :: SomeException) ->
                            do warnA ("Unable to start " ++ mp3Tool ++ ": " ++ show e)
                               pure Nothing)
 
-        flip (maybe (threadDelay 1_000_000)) mv \ (hw, r, e, pid) -> do
+        whenJust mv \ (hw, r, e, pid) -> do
 
             mhw         <- newMVar hw
             mew         <- newMVar =<< newFiltHandle e
             mfilep      <- newMVar =<< newFiltHandle r
 
             modifyST $ \st ->
-                       st { mp3pid    = Just pid
-                          , writeh    = mhw
-                          , errh      = mew
-                          , readf     = mfilep
-                          , status    = Stopped
-                          , info      = Nothing
-                          , id3       = Nothing }
+               st { mp3pid    = Just pid
+                  , writeh    = mhw
+                  , errh      = mew
+                  , readf     = mfilep
+                  , status    = Stopped
+                  , info      = Nothing
+                  , id3       = Nothing
+                  }
 
             catch @SomeException (void $ waitForProcess pid) (\_ -> pure ())
             stop <- getsST doNotResuscitate
             when stop exitSuccess
             warnA $ "Restarting " ++ mppath ++ " ..."
 
+        -- Delay to slow spawn loops in case of trouble.
+        threadDelay 4_000_000
+
+
 ------------------------------------------------------------------------
 
 -- | When the editor state has been modified, refresh, then wait
@@ -238,7 +233,7 @@ showTimeDiff = showTimeDiff_ False
 
 ------------------------------------------------------------------------
 
--- | Once each half second, wake up a and redraw the clock
+-- | Once each half second, wake up and redraw the clock
 clockLoop :: IO ()
 clockLoop = runForever $ threadDelay delay >> UI.refreshClock
   where
@@ -276,10 +271,8 @@ mpgInput field = runForever $ do
 run :: IO ()
 run = runForever $ sequence_ . keymap =<< getKeys
   where
-    getKeys = unsafeInterleaveIO $ do
-            c  <- UI.getKey
-            cs <- getKeys
-            pure (c:cs) -- A lazy list of curses keys
+    -- A lazy list of curses keys
+    getKeys = unsafeInterleaveIO $ (:) <$> UI.getKey <*> getKeys
 
 ------------------------------------------------------------------------
 
@@ -287,7 +280,7 @@ run = runForever $ sequence_ . keymap =<< getKeys
 shutdown :: Maybe String -> IO ()
 shutdown ms =
     do  silentlyModifyST $ \st -> st { doNotResuscitate = True }
-        discardErrors writeSt
+        discardErrors writeState
         withST $ \st -> do
             case mp3pid st of
                 Nothing  -> pure ()
@@ -409,11 +402,6 @@ blacklist = do
     let fe = music st ! cursor st
     in P.intercalate (P.singleton '/') [dname $ folders st ! fdir fe, fbase fe]
 
-
--- | Play a random song
-playRandom :: IO ()
-playRandom = modifySTM jumpToRandom
-
 -- | Jump to a random song
 jumpToRandom :: HState -> IO HState
 jumpToRandom st = do
@@ -425,29 +413,25 @@ jumpToRandom st = do
 -- If we're at the beginning, and loop mode is on, then loop to the end
 -- If we're in random mode, play the next random track
 playPrev :: IO ()
-playPrev = do
-    md <- getsST mode
-    if md == Random then playRandom else
-      modifySTM $ \st -> do
-      let i   = current st
-      if
-        | i > 0             -> playAtN st (subtract 1)      -- just the prev track
-        | mode st == Loop   -> playAtN st (const (size st - 1))  -- maybe loop
-        | otherwise         -> pure    st            -- else stop at end
+playPrev = modifySTM \st -> case mode st of
+    Random -> jumpToRandom st
+    Single -> pure st
+    _ | current st > 0
+           -> playAtN st (subtract 1)
+    Loop   -> playAtN st (const (size st - 1))
+    Once   -> pure st
 
 -- | Play the song following the current song, if we're not at the end
 -- If we're at the end, and loop mode is on, then loop to the start
 -- If we're in random mode, play the next random track
 playNext :: IO ()
-playNext = do
-    md <- getsST mode
-    if md == Random then playRandom else
-      modifySTM $ \st -> do
-      let i   = current st
-      if
-        | i < size st - 1   -> playAtN st (+ 1)      -- just the next track
-        | mode st == Loop   -> playAtN st (const 0)  -- maybe loop
-        | otherwise         -> pure    st            -- else stop at end
+playNext = modifySTM \st -> case mode st of
+    Random -> jumpToRandom st
+    Single -> pure st
+    _ | current st < size st - 1
+           -> playAtN st (+ 1)
+    Loop   -> playAtN st (const 0)
+    Once   -> pure st
 
 -- | Generic next song selection
 -- If the cursor and current are currently the same, continue that.
@@ -604,7 +588,7 @@ showHist = do
         histVisible = Just $ do
             (tm, ix) <- toList $ playHist st
             pure (UTF8.toString $ showTimeDiff_ True tm now
-                , UTF8.toString $ fbase $ music st ! ix)
+                , (ix, UTF8.toString $ fbase $ music st ! ix))
         }
 
 -- | Toggle the mode flag
@@ -615,34 +599,27 @@ nextMode = modifyST $ \st -> st { mode = next (mode st) }
 
 ------------------------------------------------------------------------
 
-getCachePath :: IO FilePath
-getCachePath = getXdgDirectory XdgCache $ "hmp3" </> "playlist.db"
+getStatePath :: IO FilePath
+getStatePath = getXdgDirectory XdgState "hmp3"
 
--- | Saving the playlist 
--- Only save if there's something to save. Should prevent dbs being wiped
--- if curses crashes before the state is read.
-writeSt :: IO ()
-writeSt = do
-    f <- getCachePath
-    withST \st -> when (size st > 0) do
-        let arr1 = music st
-            arr2 = folders st
-            idx  = current st
-            mde  = mode st
-        createDirectoryIfMissing True $ takeDirectory f
-        writeTree f $ SerialT {
-            ser_farr = arr1
-           ,ser_darr = arr2
-           ,ser_indx = idx
-           ,ser_mode = mde
-          }
+-- | Save mode state
+writeState :: IO ()
+writeState = do
+    dir <- getStatePath
+    createDirectoryIfMissing True dir
+    mode <- getsST mode
+    writeFile (dir </> "mode") $ show mode ++ "\n"
 
--- | Read the playlist back
-readSt :: IO (Maybe SerialT)
-readSt = do
-    f <- getCachePath
+-- | Read mode state
+readState :: IO Mode
+readState = do
+    dir <- getStatePath
+    let f = dir </> "mode"
     b <- doesFileExist f
-    if b then Just <$!> readTree f else pure Nothing
+    modeM <- if b
+        then readMaybe <$!> readFile f
+        else pure Nothing
+    pure $ fromMaybe (mode emptySt) modeM
 
 ------------------------------------------------------------------------
 -- Read styles from style.conf
@@ -658,17 +635,14 @@ loadConfig = do
     if b then do
         str' <- readFile f
         str <- let (old, new) = ("hmp3_helpscreen", "hmp3_modals") in
-            if old `isInfixOf` str'
-            then do
-                warnA $ old ++ " is now " ++ new ++ " in style.conf"
-                let (ix, rest) = head $ filter (\ (_, s) -> old `isPrefixOf` s) $ zip [0..] $ tails str'
-                pure $ take ix str' ++ new ++ drop (length old) rest
-            else pure str'
-        msty <- catch (fmap Just $ evaluate $ read str)
-                      (\ (_ :: SomeException) ->
-                        warnA "Parse error in style.conf" $> Nothing)
-        case msty of
-            Nothing  -> pure ()
+            case findIndex (old `isPrefixOf`) $ tails str' of
+                Just ix -> do
+                    warnA $ old ++ " is now " ++ new ++ " in style.conf"
+                    pure $ take ix str' ++ new ++ drop (ix + length old) str'
+                _ -> pure str'
+        case readMaybe str of
+            Nothing  -> do
+                warnA "Parse error in style.conf"
             Just rsty -> do
                 let sty = buildStyle rsty
                 initcolours sty
